type ftm;;

val mk_db : int -> ftm
val mk_name : int -> ftm
val mk_norm_ap : ftm -> ftm -> ftm
val mk_norm_lam : int -> ftm -> ftm
val mk_false : ftm  (* The unique copy *)
val mk_imp : ftm -> ftm -> ftm
val mk_all : int -> ftm -> ftm
val mk_choice : int -> ftm
val mk_eq : int -> ftm -> ftm -> ftm

(* These 2 do not beta-eta-normalize. Use only is sure.
   Otherwise they are stored in caches and will cause problems! *)
val mk_unsafe_ap : ftm -> ftm -> ftm
val mk_unsafe_lam : int -> ftm -> ftm

(* Explicit requests for substitutions *)
val uptrm : ftm -> int -> int -> ftm
val subst : ftm -> int -> ftm -> ftm

(* FT_None should never happen. It is there to avoid shifting indices by one
   when extracting the index of the C structure *)
type ftm_tag =
    FT_None
  | FT_DB
  | FT_Name
  | FT_Ap
  | FT_Lam
  | FT_False
  | FT_Imp
  | FT_All
  | FT_Choice
  | FT_Eq
val get_tag : ftm -> ftm_tag

(* get_no lets you get the number of the type or of the name stored under the tag *)
val get_no : ftm -> int
val get_l : ftm -> ftm
val get_r : ftm -> ftm
val get_head : ftm -> ftm
val get_head_spine : ftm -> (ftm * ftm list)

val get_maxv : ftm -> int
val get_fv0 : ftm -> int -> bool
val get_fv0_0 : ftm -> bool
val print : ftm -> unit

(* inefficient - only there for debugging *)
val size : ftm -> int

type hashtbli1
val hashtbli1_make : int -> hashtbli1
val hashtbli1_add : hashtbli1 -> ftm -> int -> unit
val hashtbli1_find : hashtbli1 -> ftm -> int
val hashtbli1_mem : hashtbli1 -> ftm -> bool

type vectori
val vectori_make : int -> vectori
val vectori_add : vectori -> int -> ftm -> unit
val vectori_find : vectori -> int -> ftm
